"""Configuration of lights."""

import collections.abc
import dataclasses
import logging
import typing

import HABApp.util

import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)

BrightnessTypes = typing.Union[list[typing.Union[float, bool]], float, bool]


@dataclasses.dataclass
class BrightnessTimeout:
	"""Define brightness and timeout for light states."""
	brightness: int | bool
	timeout: float

	def __post_init__(self):
		"""Check if all values where set correct.

		:raises habapp_rules.common.exceptions.HabAppRulesConfigurationException: if config is not valid."""
		if self.brightness is False:
			# Default if the light should be switched off e.g. for leaving / sleeping
			if not self.timeout:
				self.timeout = 0.5
			return

		if not self.brightness or not self.timeout:
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"Brightness and timeout are not valid: brightness = {self.brightness} | timeout = {self.timeout}")


@dataclasses.dataclass
class FunctionConfig:
	"""Define brightness and timeout values for one function."""
	day: BrightnessTimeout | None
	night: BrightnessTimeout | None
	sleeping: BrightnessTimeout | None


@dataclasses.dataclass
class LightConfig:
	"""Configuration for basic lights"""
	on: FunctionConfig  # pylint: disable=invalid-name
	pre_off: FunctionConfig | None
	leaving: FunctionConfig | None
	pre_sleep: FunctionConfig | None
	pre_sleep_prevent: collections.abc.Callable[[], bool] | HABApp.openhab.items.OpenhabItem | None = None

	def __post_init__(self):
		"""Check if light config is correct.

		:raises habapp_rules.common.exceptions.HabAppRulesConfigurationException: if config is not correct."""
		if not self.on or not all(dataclasses.asdict(self.on).values()):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("For function 'on' all brightness / timeout values must be set.")

		if getattr(self.pre_sleep, "sleeping", None):
			LOGGER.warning("It's not allowed to set brightness / timeout for pre_sleep.sleeping. Set it to None")
			self.pre_sleep.sleeping = None


@dataclasses.dataclass
class LightConfigExtended(LightConfig):
	"""Configuration for extended lights."""
	motion: FunctionConfig | None = None
	door: FunctionConfig | None = None
	off_at_door_closed_during_leaving: bool = False  # this can be used to switch lights off, when door is closed in leaving state
	hand_off_lock_time: int = 20


CONFIG_DEFAULT = LightConfig(
	on=FunctionConfig(day=BrightnessTimeout(True, 600), night=BrightnessTimeout(80, 180), sleeping=BrightnessTimeout(20, 60)),
	pre_off=FunctionConfig(day=BrightnessTimeout(50, 10), night=BrightnessTimeout(40, 7), sleeping=BrightnessTimeout(10, 7)),
	leaving=FunctionConfig(day=BrightnessTimeout(False, 0), night=BrightnessTimeout(False, 0), sleeping=BrightnessTimeout(False, 0)),
	pre_sleep=FunctionConfig(day=BrightnessTimeout(False, 10), night=BrightnessTimeout(False, 10), sleeping=None),
)

CONFIG_DEFAULT_EXTENDED = LightConfigExtended(
	on=FunctionConfig(day=BrightnessTimeout(True, 600), night=BrightnessTimeout(80, 180), sleeping=BrightnessTimeout(20, 60)),
	pre_off=FunctionConfig(day=BrightnessTimeout(50, 10), night=BrightnessTimeout(40, 7), sleeping=BrightnessTimeout(10, 7)),
	leaving=FunctionConfig(day=BrightnessTimeout(False, 0), night=BrightnessTimeout(False, 0), sleeping=BrightnessTimeout(False, 0)),
	pre_sleep=FunctionConfig(day=BrightnessTimeout(False, 10), night=BrightnessTimeout(False, 10), sleeping=None),
	motion=FunctionConfig(day=BrightnessTimeout(True, 600), night=BrightnessTimeout(80, 180), sleeping=BrightnessTimeout(20, 60)),
	door=FunctionConfig(day=BrightnessTimeout(True, 60), night=BrightnessTimeout(80, 30), sleeping=None),
	off_at_door_closed_during_leaving=False
)



"""Implementations of logical functions."""
import abc
import logging

import HABApp

import habapp_rules.core.helper
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class _Base(HABApp.Rule):
	"""Base class for logical functions."""

	def __init__(self, input_names: list[str], output_name: str) -> None:
		"""Init a logical function.

		:param input_names: list of input items (must be either Switch or Contact and all have to match to output_item)
		:param output_name: name of output item
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, output_name)

		self._output_item = HABApp.openhab.items.OpenhabItem.get_item(output_name)

		if isinstance(self._output_item, HABApp.openhab.items.SwitchItem):
			self._positive_state = "ON"
			self._negative_state = "OFF"
		elif isinstance(self._output_item, HABApp.openhab.items.ContactItem):
			self._positive_state = "CLOSED"
			self._negative_state = "OPEN"
		else:
			raise TypeError(f"Item type '{type(self._output_item)}' is not supported. Type must be SwitchItem or ContactItem")

		self._input_items = []
		for name in input_names:
			if isinstance(input_item := HABApp.openhab.items.OpenhabItem.get_item(name), type(self._output_item)):
				self._input_items.append(input_item)
				input_item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateEventFilter())
			else:
				self._instance_logger.error(f"Item '{name}' must have the same type like the output item. Expected: {type(self._output_item)} | actual : {type(input_item)}")

		self._cb_input_event(None)

	@abc.abstractmethod
	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""

	def _set_output_sate(self, output_state: str) -> None:
		"""Set state to the output element

		:param output_state: state which will be set
		"""
		if isinstance(self._output_item, HABApp.openhab.items.ContactItem):
			self._output_item.oh_post_update(output_state)
		else:
			habapp_rules.core.helper.send_if_different(self._output_item.name, output_state)


class And(_Base):
	"""Logical AND function."""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if all(item.value == self._positive_state for item in self._input_items) else self._negative_state
		self._set_output_sate(output_state)


class Or(_Base):
	"""Logical OR function."""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if any(item.value == self._positive_state for item in self._input_items) else self._negative_state
		self._set_output_sate(output_state)



"""Test for timeout_list"""
import dataclasses
import time
import typing


@dataclasses.dataclass
class ValueWithTimeout:
	"""Define item for TimeoutList"""
	value: typing.Any
	timeout: float
	add_timestamp: float


class TimeoutList:
	"""List like class, where every item has a timeout, which will remove it from the list."""

	def __init__(self) -> None:
		"""Init class"""
		self.__items: list[ValueWithTimeout] = []

	def __repr__(self) -> str:
		"""Get representation of all list items (without timeout)

		:return: all list elements which are currently in the list
		"""
		self.__remove_old_items()
		return str([itm.value for itm in self.__items])

	def __bool__(self) -> bool:
		"""Check if list has items

		:return: true if items in list
		"""
		self.__remove_old_items()
		return bool(self.__items)

	def __contains__(self, item: typing.Any) -> bool:
		"""Check if an item is in the list

		:param item: item which should be checked
		:return: true if item is in the list
		"""
		self.__remove_old_items()
		return item in [itm.value for itm in self.__items]

	def __getitem__(self, index: int) -> typing.Any:
		"""Get item from list by index

		:param index: index of item position in list.
		:return: item from list
		"""
		self.__remove_old_items()
		return self.__items[index].value

	def __eq__(self, other: typing.Any) -> bool:
		"""Check if equal.

		:param other: other item
		:return: true if equal
		"""
		if isinstance(other, TimeoutList):
			return repr(self) == repr(other)

		if isinstance(other, list):
			self.__remove_old_items()
			return [itm.value for itm in self.__items] == other

		return False

	def __remove_old_items(self) -> None:
		"""Remove items from list, which are timed-out."""
		current_time = time.time()
		self.__items = [itm for itm in self.__items if current_time - itm.add_timestamp < itm.timeout]

	def append(self, item: typing.Any, timeout: float) -> None:
		"""Add item to list

		:param item: item which should be added to the list
		:param timeout: timeout, after which the item is not valid anymore
		"""
		self.__items.append(ValueWithTimeout(item, timeout, time.time()))

	def remove(self, item: typing.Any) -> None:
		"""Remove item from list. If there are duplicates. The first element will be removed.

		:param item: item which should be deleted
		:raises ValueError: if item not in list
		"""
		item_to_remove = next((itm for itm in self.__items if itm.value == item), None)

		if not item_to_remove:
			raise ValueError(f"{self.__class__.__name__}.remove(x): x not in list")

		self.__items.remove(item_to_remove)

	def pop(self, element_index: int) -> typing.Any:
		"""Pop item from list

		:param element_index: list index of element which should be deleted
		:return: item which was removed
		:raises IndexError: if index is out of range
		:raises TypeError: if index can not be interpreted as integer
		"""
		return self.__items.pop(element_index).value
